"""
Array API specification for element-wise functions.

https://data-apis.org/array- api/latest/API_specification/elementwise_functions.html.
"""

from rbc.stdlib import Expose, BinaryUfuncExpose, UnaryUfuncExpose, API, determine_input_type
import numpy as np
from rbc import typesystem
from rbc.omnisci_backend import ArrayPointer, Array
from numba.core import types


__all__ = [
    'add', 'subtract', 'multiply', 'divide', 'logaddexp', 'logaddexp2',
    'true_divide', 'floor_divide', 'pow', 'remainder', 'mod',
    'fmod', 'gcd', 'lcm', 'bitwise_and', 'bitwise_or', 'bitwise_xor',
    'bitwise_not', 'atan2', 'hypot',
    'greater', 'greater_equal', 'less', 'less_equal', 'not_equal',
    'equal', 'logical_and', 'logical_or', 'logical_xor', 'maximum',
    'minimum', 'fmax', 'fmin', 'nextafter', 'ldexp', 'negative',
    'positive', 'rint', 'sign', 'abs', 'conj',
    'conjugate', 'exp', 'exp2', 'log', 'log2', 'log10', 'expm1',
    'log1p', 'sqrt', 'square', 'reciprocal', 'bitwise_not', 'sin', 'cos',
    'tan', 'asin', 'acos', 'atan', 'sinh', 'cosh', 'tanh',
    'asinh', 'acosh', 'atanh', 'degrees', 'radians', 'deg2rad',
    'rad2deg', 'logical_not', 'isfinite', 'isinf', 'isnan', 'fabs',
    'floor', 'ceil', 'trunc', 'signbit', 'copysign', 'spacing',
    'heaviside', 'bitwise_left_shift', 'bitwise_right_shift',
    'round',
    # numpy specifics
    'power', 'arctan2', 'left_shift', 'right_shift', 'absolute',
    'invert', 'arcsin', 'arctan', 'arccos', 'arcsinh', 'arccosh', 'arctanh'
]


expose = Expose(globals(), 'elementwise_functions')
binary_expose = BinaryUfuncExpose(globals(), 'elementwise_functions')
unary_expose = UnaryUfuncExpose(globals(), 'elementwise_functions')


# math functions
@binary_expose.implements(np.add)
def _omnisci_add(x1, x2):
    """
    Calculates the sum for each element x1_i of the input array x1 with the respective element
    x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.subtract)
def _omnisci_ufunc_subtract(x1, x2):
    """
    Calculates the difference for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.multiply)
def _omnisci_ufunc_multiply(x1, x2):
    """
    Calculates the product for each element x1_i of the input array x1 with the respective element
    x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.divide, ufunc_name='divide')
def _omnisci_ufunc_divide(x1, x2):
    """
    Calculates the division for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.logaddexp)
def _omnisci_ufunc_logaddexp(x1, x2):
    """
    Calculates the logarithm of the sum of exponentiations ``log(exp(x1) + exp(x2))`` for each
    element x1_i of the input array x1 with the respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.copysign, api=API.NUMPY_API)
def _omnisci_ufunc_copysign(x1, x2):
    pass


@binary_expose.implements(np.logaddexp2, api=API.NUMPY_API)
def _omnisci_ufunc_logaddexp2(x1, x2):
    pass


@binary_expose.implements(np.true_divide, api=API.NUMPY_API)
def _omnisci_ufunc_true_divide(x1, x2):
    pass


@binary_expose.implements(np.floor_divide)
def _omnisci_ufunc_floor_divide(x1, x2):
    """
    Rounds the result of dividing each element x1_i of the input array x1 by the respective
    element x2_i of the input array x2 to the greatest (i.e., closest to +infinity) integer-value
    number that is not greater than the division result.
    """
    pass


@binary_expose.implements(np.power, ufunc_name='power', api=API.NUMPY_API)
def _omnisci_ufunc_power(x1, x2):
    pass


@binary_expose.implements(np.power, ufunc_name='pow')
def _omnisci_ufunc_pow(x1, x2):
    """
    Calculates an implementation-dependent approximation of exponentiation by raising each element
    x1_i (the base) of the input array x1 to the power of x2_i (the exponent), where x2_i is the
    corresponding element of the input array x2.
    """
    pass


@binary_expose.not_implemented('float_power')  # not supported by Numba
def _omnisci_ufunc_float_power(x1, x2):
    pass


@binary_expose.implements(np.remainder)
def _omnisci_ufunc_remainder(x1, x2):
    """
    Returns the remainder of division for each element x1_i of the input array x1 and the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.mod, ufunc_name='mod', api=API.NUMPY_API)
def _omnisci_ufunc_mod(x1, x2):
    pass


@binary_expose.implements(np.fmod, api=API.NUMPY_API)
def _omnisci_ufunc_fmod(x1, x2):
    pass


@binary_expose.not_implemented('divmod')  # not supported by Numba
def _omnisci_ufunc_divmod(x1, x2):
    pass


@binary_expose.implements(np.gcd, api=API.NUMPY_API)
def _omnisci_ufunc_gcd(x1, x2):
    pass


@binary_expose.implements(np.lcm, api=API.NUMPY_API)
def _omnisci_ufunc_lcm(x1, x2):
    pass


# Bit-twiddling functions
@binary_expose.implements(np.bitwise_and)
def _omnisci_ufunc_bitwise_and(x1, x2):
    """
    Computes the bitwise AND of the underlying binary representation of each element x1_iof the
    input array x1 with the respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.bitwise_or)
def _omnisci_ufunc_bitwise_or(x1, x2):
    """
    Computes the bitwise OR of the underlying binary representation of each element x1_i of the
    input array x1 with the respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.bitwise_xor)
def _omnisci_ufunc_bitwise_xor(x1, x2):
    """
    Computes the bitwise XOR of the underlying binary representation of each element x1_i of the
    input array x1 with the respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.bitwise_not, ufunc_name='bitwise_not')
def _omnisci_ufunc_bitwise_not(x1, x2):
    """
    Computes the bitwise NOR of the underlying binary representation of each element x1_i of the
    input array x1 with the respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.left_shift, api=API.NUMPY_API)
def _omnisci_ufunc_left_shift(x1, x2):
    pass


@binary_expose.implements(np.left_shift, ufunc_name='bitwise_left_shift')
def _omnisci_ufunc_bitwise_left_shift(x1, x2):
    """
    Shifts the bits of each element x1_i of the input array x1 to the left by appending x2_i
    (i.e., the respective element in the input array x2) zeros to the right of x1_i.
    """
    pass


@binary_expose.implements(np.right_shift, api=API.NUMPY_API)
def _omnisci_ufunc_right_shift(x1, x2):
    pass


@binary_expose.implements(np.right_shift, ufunc_name='bitwise_right_shift')
def _omnisci_ufunc_bitwise_right_shift(x1, x2):
    """
    Shifts the bits of each element x1_i of the input array x1 to the right by appending x2_i
    (i.e., the respective element in the input array x2) zeros to the right of x1_i.
    """
    pass


# trigonometric functions
@binary_expose.implements(np.arctan2, api=API.NUMPY_API)
def _omnisci_ufunc_arctan2(x1, x2):
    pass


@binary_expose.implements(np.arctan2, ufunc_name='atan2')
def _omnisci_ufunc_atan2(x1, x2):
    """
    Calculates an implementation-dependent approximation of the inverse tangent of the quotient
    x1/x2, having domain [-infinity, +infinity] x ``[-infinity, +infinity]`` (where the x notation
    denotes the set of ordered pairs of elements (x1_i, x2_i)) and codomain [-π, +π], for each
    pair of elements (x1_i, x2_i) of the input arrays x1 and x2, respectively.
    """
    pass


@binary_expose.implements(np.hypot, api=API.NUMPY_API)
def _omnisci_ufunc_hypot(x1, x2):
    pass


# Comparison functions
@binary_expose.implements(np.greater, dtype=typesystem.boolean8)
def _omnisci_ufunc_greater(x1, x2):
    """
    Computes the truth value of x1_i > x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.greater_equal, dtype=typesystem.boolean8)
def _omnisci_ufunc_greater_equal(x1, x2):
    """
    Computes the truth value of x1_i >= x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.less, dtype=typesystem.boolean8)
def _omnisci_ufunc_less(x1, x2):
    """
    Computes the truth value of x1_i < x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.less_equal, dtype=typesystem.boolean8)
def _omnisci_ufunc_less_equal(x1, x2):
    """
    Computes the truth value of x1_i <= x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.not_equal, dtype=typesystem.boolean8)
def _omnisci_ufunc_not_equal(x1, x2):
    """
    Computes the truth value of x1_i != x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.equal, dtype=typesystem.boolean8)
def _omnisci_ufunc_equal(x1, x2):
    """
    Computes the truth value of x1_i == x2_i for each element x1_i of the input array x1 with the
    respective element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.logical_and, dtype=typesystem.boolean8)
def _omnisci_ufunc_logical_and(x1, x2):
    """
    Computes the logical AND for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.logical_or, dtype=typesystem.boolean8)
def _omnisci_ufunc_logical_or(x1, x2):
    """
    Computes the logical OR for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.logical_xor, dtype=typesystem.boolean8)
def _omnisci_ufunc_logical_xor(x1, x2):
    """
    "Computes the logical XOR for each element x1_i of the input array x1 with the respective
    element x2_i of the input array x2.
    """
    pass


@binary_expose.implements(np.maximum, api=API.NUMPY_API)
def _omnisci_ufunc_maximum(x1, x2):
    pass


@binary_expose.implements(np.minimum, api=API.NUMPY_API)
def _omnisci_ufunc_minimum(x1, x2):
    pass


@binary_expose.implements(np.fmax, api=API.NUMPY_API)
def _omnisci_ufunc_fmax(x1, x2):
    pass


@binary_expose.implements(np.fmin, api=API.NUMPY_API)
def _omnisci_ufunc_fmin(x1, x2):
    pass


# Floating functions
@binary_expose.implements(np.nextafter, api=API.NUMPY_API)
def _omnisci_ufunc_nextafter(x1, x2):
    pass


@binary_expose.implements(np.ldexp, api=API.NUMPY_API)
def _omnisci_ufunc_ldexp(x1, x2):
    pass


##################################################################


@unary_expose.implements(np.around, ufunc_name='round')
def _omnisci_ufunc_round(a):
    """
    Rounds each element x_i of the input array x to the nearest integer-valued number.
    """
    pass


@unary_expose.implements(np.negative)
def _omnisci_ufunc_negative(a):
    """Computes the numerical negative of each element x_i (i.e., y_i = -x_i) of the
    input array x."""
    pass


@unary_expose.implements(np.positive)
def _omnisci_ufunc_positive(a):
    """Computes the numerical positive of each element x_i (i.e., y_i = +x_i) of the
    input array x."""
    pass


@unary_expose.implements(np.absolute, api=API.NUMPY_API)
def _omnisci_ufunc_absolute(a):
    pass


@unary_expose.implements(np.absolute, ufunc_name='abs')
def _omnisci_ufunc_abs(a):
    """
    Calculates the absolute value for each element x_i of the input array x (i.e., the element-
    wise result has the same magnitude as the respective element in x but has positive sign).
    """
    pass


@unary_expose.implements(np.rint, api=API.NUMPY_API)
def _omnisci_ufunc_rint(a):
    pass


@unary_expose.implements(np.sign)
def _omnisci_ufunc_sign(a):
    """
    Returns an indication of the sign of a number for each element x_i of the input array x.
    """
    pass


@unary_expose.implements(np.conj, ufunc_name='conj', api=API.NUMPY_API)
def _omnisci_ufunc_conj(a):
    pass


@unary_expose.implements(np.conjugate, api=API.NUMPY_API)
def _omnisci_ufunc_conjugate(a):
    pass


@unary_expose.implements(np.exp)
def _omnisci_ufunc_exp(a):
    """Calculates an implementation-dependent approximation to the exponential function, having
    domain.

    [-infinity, +infinity] and codomain [+0, +infinity], for each
    element x_i of the input array x (e raised to the power of x_i,
    where e is the base of the natural logarithm).

    """
    pass


@unary_expose.implements(np.exp2, api=API.NUMPY_API)
def _omnisci_ufunc_exp2(a):
    pass


@unary_expose.implements(np.log)
def _omnisci_ufunc_log(a):
    """
    Calculates an implementation-dependent approximation to the natural (base e) logarithm, having
    domain [0, +infinity] and codomain [-infinity, +infinity], for each element x_i of the input
    array x.
    """
    pass


@unary_expose.implements(np.log2)
def _omnisci_ufunc_log2(a):
    """Calculates an implementation-dependent approximation to the base 2 logarithm, having
    domain.

    [0,

    +infinity] and codomain [-infinity, +infinity], for each element x_i
    of the input array x.

    """
    pass


@unary_expose.implements(np.log10)
def _omnisci_ufunc_log10(a):
    """Calculates an implementation-dependent approximation to the base 10 logarithm, having
    domain.

    [0,

    +infinity] and codomain [-infinity, +infinity], for each element x_i
    of the input array x.

    """
    pass


@unary_expose.implements(np.expm1)
def _omnisci_ufunc_expm1(a):
    """Calculates an implementation-dependent approximation to exp(x)-1, having domain [-infinity,

    +infinity] and codomain [-1, +infinity], for each element x_i of the
    input array x.

    """
    pass


@unary_expose.implements(np.log1p)
def _omnisci_ufunc_log1p(a):
    """
    Calculates an implementation-dependent approximation to log(1+x), where log refers to the
    natural (base e) logarithm, having domain [-1, +infinity] and codomain [-infinity, +infinity],
    for each element x_i of the input array x.
    """
    pass


@unary_expose.implements(np.sqrt)
def _omnisci_ufunc_sqrt(a):
    """
    Calculates the square root, having domain [0, +infinity] and codomain [0, +infinity], for each
    element x_i of the input array x.
    """
    pass


@unary_expose.implements(np.square)
def _omnisci_ufunc_square(a):
    """Squares (x_i * x_i) each element x_i of the input array x."""
    pass


# @unary_expose.implements(np.cbrt)  # not supported by numba
@unary_expose.not_implemented('cbrt')
def _omnisci_ufunc_cbrt(a):
    pass


@unary_expose.implements(np.reciprocal, api=API.NUMPY_API)
def _omnisci_ufunc_reciprocal(a):
    pass


# Bit-twiddling functions
@unary_expose.implements(np.invert, api=API.NUMPY_API)
def _omnisci_ufunc_invert(a):
    pass


@unary_expose.implements(np.invert, ufunc_name='bitwise_invert')
def _omnisci_ufunc_bitwise_invert(a):
    """
    Inverts (flips) each bit for each element x_i of the input array x.
    """
    pass


# trigonometric functions
@unary_expose.implements(np.sin)
def _omnisci_ufunc_sin(a):
    """Calculates an implementation-dependent approximation to the sine, having domain (-infinity,

    +infinity) and codomain [-1, +1], for each element x_i of the input
    array x.

    """
    pass


@unary_expose.implements(np.cos)
def _omnisci_ufunc_cos(a):
    """Calculates an implementation-dependent approximation to the cosine, having domain
    (-infinity,

    +infinity) and codomain [-1, +1], for each element x_i of the input
    array x.

    """
    pass


@unary_expose.implements(np.tan)
def _omnisci_ufunc_tan(a):
    """Calculates an implementation-dependent approximation to the tangent, having domain
    (-infinity,

    +infinity) and codomain (-infinity, +infinity), for each element x_i
    of the input array x.

    """
    pass


@unary_expose.implements(np.arcsin, api=API.NUMPY_API)
def _omnisci_ufunc_arcsin(a):
    pass


@unary_expose.implements(np.arcsin, ufunc_name='asin')
def _omnisci_ufunc_asin(a):
    """
    Calculates an implementation-dependent approximation of the principal value of the inverse
    sine, having domain [-1, +1] and codomain [-π/2, +π/2] for each element x_i of the input array
    x.
    """
    pass


@unary_expose.implements(np.arccos, api=API.NUMPY_API)
def _omnisci_ufunc_arccos(a):
    pass


@unary_expose.implements(np.arccos, ufunc_name='acos')
def _omnisci_ufunc_acos(a):
    """
    Calculates an implementation-dependent approximation of the principal value of the inverse
    cosine, having domain [-1, +1] and codomain [+0, +π], for each element x_i of the input array
    x.
    """
    pass


@unary_expose.implements(np.arctan, api=API.NUMPY_API)
def _omnisci_ufunc_arctan(a):
    pass


@unary_expose.implements(np.arctan, ufunc_name='atan')
def _omnisci_ufunc_atan(a):
    """
    Calculates an implementation-dependent approximation of the principal value of the inverse
    tangent, having domain [-infinity, +infinity] and codomain [-π/2, +π/2], for each element x_i
    of the input array x.
    """
    pass


@unary_expose.implements(np.sinh)
def _omnisci_ufunc_sinh(a):
    """Calculates an implementation-dependent approximation to the hyperbolic sine, having domain.

    [-infinity, +infinity] and codomain [-infinity, +infinity], for each
    element x_i of the input array x.

    """
    pass


@unary_expose.implements(np.cosh)
def _omnisci_ufunc_cosh(a):
    """Calculates an implementation-dependent approximation to the hyperbolic cosine, having
    domain.

    [-infinity, +infinity] and codomain [-infinity, +infinity], for each
    element x_i in the input array x.

    """
    pass


@unary_expose.implements(np.tanh)
def _omnisci_ufunc_tanh(a):
    """Calculates an implementation-dependent approximation to the hyperbolic tangent, having
    domain.

    [-infinity, +infinity] and codomain [-1, +1], for each element x_i
    of the input array x.

    """
    pass


@unary_expose.implements(np.arcsinh, api=API.NUMPY_API)
def _omnisci_ufunc_arcsinh(a):
    pass


@unary_expose.implements(np.arcsinh, ufunc_name='asinh')
def _omnisci_ufunc_asinh(a):
    """
    Calculates an implementation-dependent approximation to the inverse hyperbolic sine, having
    domain [-infinity, +infinity] and codomain [-infinity, +infinity], for each element x_i in the
    input array x.
    """
    pass


@unary_expose.implements(np.arccosh, api=API.NUMPY_API)
def _omnisci_ufunc_arccosh(a):
    pass


@unary_expose.implements(np.arccosh, ufunc_name='acosh')
def _omnisci_ufunc_acosh(a):
    """
    Calculates an implementation-dependent approximation to the inverse hyperbolic cosine, having
    domain [+1, +infinity] and codomain [+0, +infinity], for each element x_i of the input array
    x.
    """
    pass


@unary_expose.implements(np.arctanh, api=API.NUMPY_API)
def _omnisci_ufunc_arctanh(a):
    pass


@unary_expose.implements(np.arctanh, ufunc_name='atanh')
def _omnisci_ufunc_atanh(a):
    """
    Calculates an implementation-dependent approximation to the inverse hyperbolic tangent, having
    domain [-1, +1] and codomain [-infinity, +infinity], for each element x_i of the input array
    x.
    """
    pass


@unary_expose.implements(np.degrees, api=API.NUMPY_API)
def _omnisci_ufunc_degrees(a):
    pass


@unary_expose.implements(np.radians, api=API.NUMPY_API)
def _omnisci_ufunc_radians(a):
    pass


@unary_expose.implements(np.deg2rad, api=API.NUMPY_API)
def _omnisci_ufunc_deg2rad(a):
    pass


@unary_expose.implements(np.rad2deg, api=API.NUMPY_API)
def _omnisci_ufunc_rad2deg(a):
    pass


# Comparison functions
@unary_expose.implements(np.logical_not, dtype=typesystem.boolean8)
def _omnisci_ufunc_logical_not(a):
    """
    Computes the logical NOT for each element x_i of the input array x.
    """
    pass


# Floating functions
@unary_expose.implements(np.isfinite, dtype=typesystem.boolean8)
def _omnisci_ufunc_isfinite(a):
    """
    Tests each element x_i of the input array x to determine if finite (i.e., not NaN and not
    equal to positive or negative infinity).
    """
    pass


@unary_expose.implements(np.isinf, dtype=typesystem.boolean8)
def _omnisci_ufunc_isinf(a):
    """
    Tests each element x_i of the input array x to determine if equal to positive or negative
    infinity.
    """
    pass


@unary_expose.implements(np.isnan, dtype=typesystem.boolean8)
def _omnisci_ufunc_isnan(a):
    """
    Tests each element x_i of the input array x to determine whether the element is NaN.
    """
    pass


@unary_expose.implements(np.fabs, dtype=types.double, api=API.NUMPY_API)
def _omnisci_ufunc_fabs(a):
    pass


@unary_expose.implements(np.floor, dtype=types.double)
def _omnisci_ufunc_floor(a):
    """
    Rounds each element x_i of the input array x to the greatest (i.e., closest to +infinity)
    integer-valued number that is not greater than x_i.
    """
    pass


@unary_expose.implements(np.ceil, dtype=types.double)
def _omnisci_ufunc_ceil(a):
    """
    Rounds each element x_i of the input array x to the smallest (i.e., closest to -infinity)
    integer-valued number that is not less than x_i.
    """
    pass


@unary_expose.implements(np.trunc, dtype=types.double)
def _omnisci_ufunc_trunc(a):
    """
    Rounds each element x_i of the input array x to the integer-valued number that is closest to
    but no greater than x_i.
    """
    pass


# not supported?
# @unary_expose.implements(np.isnat, dtype=types.int8)
@unary_expose.not_implemented('isnat')
def _omnisci_ufunc_isnat(a):
    pass


# issue 152:
@unary_expose.implements(np.signbit, dtype=typesystem.boolean8, api=API.NUMPY_API)
def _omnisci_ufunc_signbit(a):
    pass


@unary_expose.implements(np.spacing, dtype=types.double, api=API.NUMPY_API)
def _omnisci_ufunc_spacing(a):
    pass


@expose.implements('heaviside', api=API.NUMPY_API)
def _impl_heaviside(x1, x2):
    nb_dtype = types.double
    typA = determine_input_type(x1)
    typB = determine_input_type(x2)
    if isinstance(x1, ArrayPointer):
        def impl(x1, x2):
            sz = len(x1)
            r = Array(sz, nb_dtype)
            for i in range(sz):
                r[i] = heaviside(x1[i], x2)  # noqa: F821
            return r
        return impl
    else:
        def impl(x1, x2):
            if typA(x1) < 0:
                return nb_dtype(0)
            elif typA(x1) == 0:
                return nb_dtype(typB(x2))
            else:
                return nb_dtype(1)
        return impl



__all__ = ['OmnisciColumnListType']


from numba.core import extending
from rbc.typesystem import Type
from rbc import structure_type


class OmnisciColumnListType(Type):
    """OmnisciColumnListType<T> is a typesystem custom type that
    represents a pointer type of the following struct type:

      {
        T** ptrs
        int64_t length
        int64_t size
      }

    """

    @property
    def __typesystem_type__(self):
        element_type = self[0][0]
        ptrs_t = element_type.pointer().pointer().params(name='ptrs')
        length_t = Type.fromstring('int64 length')
        size_t = Type.fromstring('int64 size')
        return Type(ptrs_t, length_t, size_t).params(
            NumbaPointerType=OmnisciColumnListNumbaType).pointer()


class OmnisciColumnListNumbaType(structure_type.StructureNumbaPointerType):
    def get_getitem_impl(self):
        def impl(x, i):
            return x.ptrs[i]
        return impl


@extending.overload_attribute(OmnisciColumnListNumbaType, 'nrows')
def get_nrows(clst):
    def impl(clst):
        return clst.size
    return impl


@extending.overload_attribute(OmnisciColumnListNumbaType, 'ncols')
def get_ncols(clst):
    def impl(clst):
        return clst.length
    return impl



from .omnisci_array import *  # noqa: F401, F403
from .omnisci_column import *  # noqa: F401, F403
from .omnisci_bytes import *  # noqa: F401, F403
from .omnisci_metatype import *  # noqa: F401, F403
from .omnisci_text_encoding import *  # noqa: F401, F403
from .omnisci_pipeline import *  # noqa: F401, F403
from .omnisci_column_list import *  # noqa: F401, F403
from .omnisci_table_function_manager import *  # noqa: F401, F403

import rbc.omnisci_backend.mathimpl as math  # noqa: F401
import rbc.omnisci_backend.npyimpl as np  # noqa: F401
import rbc.omnisci_backend.python_operators as operators  # noqa: F401

# initialize the array api
from rbc.stdlib import array_api  # noqa: F401

__all__ = [s for s in dir() if not s.startswith('_')]



import functools
from enum import Enum
from numba.core import extending
from rbc.omnisci_backend import Array, ArrayPointer
from rbc import typesystem


ARRAY_API_ADDRESS = ("https://data-apis.org/array-api/latest/API_specification"
                     "/generated/signatures.{0}.{1}.html"
                     "#signatures.{0}.{1}")
NUMPY_API_ADDRESS = ("https://numpy.org/doc/stable/reference/generated/numpy.{0}.html")
ADDRESS = ARRAY_API_ADDRESS


class API(Enum):
    NUMPY_API = 0
    ARRAY_API = 1


def determine_dtype(a, dtype):
    if isinstance(a, ArrayPointer):
        return a.eltype if dtype is None else dtype
    else:
        return a if dtype is None else dtype


def determine_input_type(argty):
    if isinstance(argty, ArrayPointer):
        return determine_input_type(argty.eltype)

    if argty == typesystem.boolean8:
        return bool
    else:
        return argty


class Expose:
    def __init__(self, globals, module_name):
        self._globals = globals
        self.module_name = module_name

    def create_function(self, func_name):
        s = f'def {func_name}(*args, **kwargs): pass'
        exec(s, self._globals)
        fn = self._globals.get(func_name)
        return fn

    def format_docstring(self, ov_func, func_name, api):
        original_docstring = ov_func.__doc__
        if api == API.NUMPY_API:
            # Numpy
            link = (
                f"`NumPy '{func_name}' "
                f"doc <{NUMPY_API_ADDRESS.format(func_name)}>`_")
        else:
            # Array API
            link = (
                f"`Array-API '{func_name}' "
                f"doc <{ARRAY_API_ADDRESS.format(self.module_name, func_name)}>`_")

        if original_docstring is not None:
            new_doctring = f"{original_docstring}\n\n{link}"
        else:
            new_doctring = link
        return new_doctring

    def implements(self, func_name, api=API.ARRAY_API):
        fn = self.create_function(func_name)
        decorate = extending.overload(fn)

        def wrapper(overload_func):
            overload_func.__doc__ = self.format_docstring(overload_func, func_name, api)
            functools.update_wrapper(fn, overload_func)
            return decorate(overload_func)

        return wrapper

    def not_implemented(self, func_name):
        s = f'def {func_name}(*args, **kwargs): pass'
        exec(s, self._globals)

        fn = self._globals.get(func_name)

        def wraps(func):
            func.__doc__ = "❌ Not implemented"
            functools.update_wrapper(fn, func)
            return func
        return wraps


class BinaryUfuncExpose(Expose):

    def implements(self, ufunc, ufunc_name=None, dtype=None, api=API.ARRAY_API):
        """
        Wrapper for binary ufuncs that returns an array
        """
        if ufunc_name is None:
            ufunc_name = ufunc.__name__

        def binary_ufunc_impl(a, b):
            typA = determine_input_type(a)
            typB = determine_input_type(b)

            # XXX: raise error if len(a) != len(b)
            @extending.register_jitable(_nrt=False)
            def binary_impl(a, b, nb_dtype):
                sz = len(a)
                x = Array(sz, nb_dtype)
                for i in range(sz):
                    cast_a = typA(a[i])
                    cast_b = typB(b[i])
                    x[i] = nb_dtype(ufunc(cast_a, cast_b))
                return x

            @extending.register_jitable(_nrt=False)
            def broadcast(e, sz, dtype):
                b = Array(sz, dtype)
                b.fill(e)
                return b

            if isinstance(a, ArrayPointer) and isinstance(b, ArrayPointer):
                nb_dtype = determine_dtype(a, dtype)

                def impl(a, b):
                    return binary_impl(a, b, nb_dtype)
                return impl
            elif isinstance(a, ArrayPointer):
                nb_dtype = determine_dtype(a, dtype)
                other_dtype = b

                def impl(a, b):
                    b = broadcast(b, len(a), other_dtype)
                    return binary_impl(a, b, nb_dtype)
                return impl
            elif isinstance(b, ArrayPointer):
                nb_dtype = determine_dtype(b, dtype)
                other_dtype = a

                def impl(a, b):
                    a = broadcast(a, len(b), other_dtype)
                    return binary_impl(a, b, nb_dtype)
                return impl
            else:
                nb_dtype = determine_dtype(a, dtype)

                def impl(a, b):
                    cast_a = typA(a)
                    cast_b = typB(b)
                    return nb_dtype(ufunc(cast_a, cast_b))
                return impl

        fn = self.create_function(ufunc_name)

        def wrapper(overload_func):
            overload_func.__doc__ = self.format_docstring(overload_func, ufunc_name, api)
            functools.update_wrapper(fn, overload_func)

            decorate = extending.overload(fn)
            return decorate(binary_ufunc_impl)

        return wrapper


class UnaryUfuncExpose(BinaryUfuncExpose):

    def implements(self, ufunc, ufunc_name=None, dtype=None, api=API.ARRAY_API):
        """
        Wrapper for unary ufuncs that returns an array
        """
        if ufunc_name is None:
            ufunc_name = ufunc.__name__

        def unary_ufunc_impl(a):
            nb_dtype = determine_dtype(a, dtype)
            typ = determine_input_type(a)

            if isinstance(a, ArrayPointer):
                def impl(a):
                    sz = len(a)
                    x = Array(sz, nb_dtype)
                    for i in range(sz):
                        # Convert the value to type "typ"
                        cast = typ(a[i])
                        x[i] = nb_dtype(ufunc(cast))
                    return x
                return impl
            else:
                def impl(a):
                    # Convert the value to type typ
                    cast = typ(a)
                    return nb_dtype(ufunc(cast))
                return impl

        fn = self.create_function(ufunc_name)

        def wrapper(overload_func):
            overload_func.__doc__ = self.format_docstring(overload_func, ufunc_name, api)
            functools.update_wrapper(fn, overload_func)

            decorate = extending.overload(fn)
            return decorate(unary_ufunc_impl)

        return wrapper



__all__ = ['Compiler']

import os
import re
import tempfile
import subprocess
import shutil
import warnings


def run(cmd, *args):
    new_args = [cmd]
    for a in args:
        if isinstance(a, str):
            new_args.append(a)
        elif isinstance(a, (tuple, list)):
            new_args.extend(a)
        else:
            raise TypeError(f'expected str|list|tuple argument, got {type(a)}')
    r = subprocess.run(new_args, capture_output=True)
    return -r.returncode, r.stdout.decode("utf-8") + '\n' + r.stderr.decode("utf-8")


class Compiler:
    """Compilers that can generate LLVM IR
    """

    def __init__(self, compiler_exe, flags, suffix, version):
        assert isinstance(compiler_exe, str)
        assert isinstance(flags, list)
        assert isinstance(suffix, str)
        assert isinstance(version, tuple)
        self.compiler_exe = compiler_exe
        self.flags = flags
        self.suffix = suffix
        self.version = version

    def __repr__(self):
        return (f'{type(self).__name__}({self.compiler_exe!r}, {self.flags!r},'
                f' {self.suffix!r}, {self.version!r})')

    @staticmethod
    def _get_compilers(_cache=[]):
        if not _cache:
            # TODO: cuda support
            for xmode, clang, suffix in [('c++', 'clang++', '.cpp'), ('c', 'clang', '.c')]:
                exe = shutil.which(clang)
                if exe is not None:
                    s, o = run(exe, '--version')
                    m = re.search(r'(\d+[.]\d+[.]\d+)([^\s]*)', o)
                    if m is None:
                        warnings.warn(f'Failed to find version from output {o}')
                        continue
                    version = tuple(map(int, m.group(1).split('.'))) + (m.group(2),)
                    lang_stds = []
                    s, o = run(exe, '-x', xmode, '-std=', '-E', '-')
                    for line in o.splitlines():
                        if line.startswith('note: use'):
                            lang_stds.extend(re.findall(r"[']([\w\d][\w\d:+]*)[']",
                                                        line.split('for', 1)[0]))
                    flags = ['-x', xmode]
                    _cache.append((exe, version, lang_stds, flags, suffix))
                else:
                    _cache.append(None)

        for data in _cache:
            if data is not None:
                yield data

    @classmethod
    def get(cls, std='C++'):
        """Construct a compiler instance using the desired language standard.

        Returns None when no compiler is found in the
        environment. Otherwise, return Compiler instance with flags
        supporting the desired language standard.
        """
        std = std.lower()
        std = {'c': 'c11', 'c++': 'c++11', 'cxx': 'c++11'}.get(std, std)
        flags = ['-c', '-O0', '-S', '-emit-llvm']
        all_stds = []
        for exe, version, lang_stds, c_flags, suffix in cls._get_compilers():
            all_stds.extend(lang_stds)
            if std in lang_stds:
                return cls(exe, c_flags + ['-std='+std] + flags, suffix, version)
        if all_stds:
            raise ValueError(f'Language standard {std} not supported.'
                             f' Supported standards: {", ".join(all_stds)}')
        return None

    def _compile(self, source, output, flags=[]):
        s, o = run(self.compiler_exe, self.flags, flags, source, '-o', output)
        if s == 0:
            f = open(output)
            out = f.read()
            f.close()
            return out
        raise RuntimeError(f'compilation failed:\n{o}')

    def __call__(self, source, flags=[]):
        """Compile the source code and return LLVM IR string.
        """
        f = tempfile.NamedTemporaryFile(mode='w', suffix=self.suffix, delete=False)
        f.write(source)
        f.close()
        fout = tempfile.NamedTemporaryFile(mode='r', suffix='.ll', delete=False)
        fout.close()
        try:
            out = self._compile(f.name, fout.name, flags=flags)
        finally:
            os.unlink(f.name)
            os.unlink(fout.name)
        # Eliminate clang inserted attributes that may cause problems
        # when server is parsing the LLVM IR string.
        # TODO: check if the attributes can be removed in irtools
        out = out.replace(') #0 {', ') {')
        return out




import os
from collections import defaultdict
import pytest
import numpy as np


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    # TODO: use omnisci_fixture from rbc/tests/__init__.py
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = os.path.splitext(os.path.basename(__file__))[0]

    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')

    sqltypes = ['FLOAT', 'DOUBLE', 'TINYINT', 'SMALLINT', 'INT', 'BIGINT',
                'BOOLEAN', 'DOUBLE']
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b', 'd']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});')

    data = defaultdict(list)
    for i in range(5):
        for j, n in enumerate(colnames):
            if n == 'b':
                data[n].append(i % 2 == 0)
            elif n == 'd':
                data[n].append(i + 1.5)
            else:
                data[n].append(i)

    m.load_table_columnar(table_name, **data)
    m.table_name = table_name
    yield m
    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')


def test_sizer_row_multiplier_orig(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<double>)')
    def my_row_copier_mul(x, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * 2
        return m * input_row_count

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_mul(cursor(select f8 '
        f'from {omnisci.table_name}), 2));')

    for i, r in enumerate(result):
        assert r == (float((i % 5) * 2),)


def test_sizer_row_multiplier_param1(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, double, int32, RowMultiplier,'
             'OutputColumn<double>)')
    def my_row_copier_mul_param1(x, alpha, beta, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * alpha + beta
        return m * input_row_count

    alpha = 3

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_mul_param1('
        f'cursor(select f8 from {omnisci.table_name}),'
        f'cast({alpha} as double),'
        'cast(4 as int), 2));')

    for i, r in enumerate(result):
        assert r == ((i % 5) * alpha + 4,)


def test_sizer_row_multiplier_param2(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(double, Column<double>, int32, RowMultiplier,'
             'OutputColumn<double>)')
    def my_row_copier_mul_param2(alpha, x, beta, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = x[i] * alpha + beta
        return m * input_row_count

    alpha = 3

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_mul_param2('
        f'cast({alpha} as double),'
        f'cursor(select f8 from {omnisci.table_name}),'
        'cast(4 as int), 2));')

    for i, r in enumerate(result):
        assert r == ((i % 5) * alpha + 4,)


def test_sizer_constant_parameter(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, ConstantParameter, OutputColumn<double>)')
    def my_row_copier_cp(x, m, y):
        n = len(x)
        for i in range(m):
            j = i % n
            y[i] = x[j] * 2
        return m

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_cp(cursor(select f8 '
        f'from {omnisci.table_name}), 3));')
    result = list(result)
    assert len(result) == 3
    for i, r in enumerate(result):
        assert r == (i * 2,)

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_cp(cursor(select f8 '
        f'from {omnisci.table_name}), 8));')
    result = list(result)
    assert len(result) == 8
    for i, r in enumerate(result):
        assert r == ((i % 5) * 2,)


def test_sizer_return_size(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, OutputColumn<double>)')
    def my_row_copier_c(x, y):
        for i in range(13):
            j = i % len(x)
            y[i] = x[j] * 2
        return 13

    descr, result = omnisci.sql_execute(
        'select * from table(my_row_copier_c(cursor(select f8 '
        f'from {omnisci.table_name})));')
    result = list(result)
    assert len(result) == 13
    for i, r in enumerate(result):
        assert r == ((i % 5) * 2,), repr((i, r))


def test_rowmul_add_columns(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, Column<double>, double,'
             ' RowMultiplier, OutputColumn<double>)')
    def add_columns(x, y, alpha, m, r):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                r[j] = x[i] + alpha * float(y[i])
        return m * input_row_count

    alpha = 2.5

    descr, result = omnisci.sql_execute(
        'select * from table(add_columns('
        f'cursor(select f8 from {omnisci.table_name}),'
        f' cursor(select d from {omnisci.table_name}),'
        f' cast({alpha} as double), 1));')

    for i, r in enumerate(result):
        assert r == (i + alpha * (i + 1.5),)


def test_rowmul_return_mixed_columns(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    @omnisci('int32(Column<double>, RowMultiplier,'
             ' OutputColumn<double>, OutputColumn<float>)')
    def ret_mixed_columns(x, m, y, z):
        input_row_count = len(x)
        for i in range(input_row_count):
            for c in range(m):
                j = i + c * input_row_count
                y[j] = 2 * x[i]
                z[j] = np.float32(3 * x[i])
        return m * input_row_count

    descr, result = omnisci.sql_execute(
        'select * from table(ret_mixed_columns('
        f'cursor(select f8 from {omnisci.table_name}), 1));')

    for i, r in enumerate(result):
        assert r[0] == float(2 * i)
        assert r[1] == float(3 * i)

    descr, result = omnisci.sql_execute(
        'select * from table(ret_mixed_columns('
        f'cursor(select f8 from {omnisci.table_name}), 2));')

    for i, r in enumerate(result):
        assert r[0] == float(2 * (i % 5))
        assert r[1] == float(3 * (i % 5))

    descr, result = omnisci.sql_execute(
        'select out1, out0 from table(ret_mixed_columns('
        f'cursor(select f8 from {omnisci.table_name}), 1));')

    for i, r in enumerate(result):
        assert r[1] == float(2 * i)
        assert r[0] == float(3 * i)


@pytest.mark.parametrize("max_n", [-1, 3])
@pytest.mark.parametrize("sizer", [1, 2])
@pytest.mark.parametrize("num_columns", [1, 2, 3, 4])
def test_rowmul_return_multi_columns(omnisci, num_columns, sizer, max_n):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    if num_columns == 1:
        @omnisci('int32(int32, Column<double>, RowMultiplier,'
                 ' OutputColumn<double>)')
        def ret_1_columns(n, x1, m, y1):
            input_row_count = len(x1)
            for i in range(input_row_count):
                for c in range(m):
                    j = i + c * input_row_count
                    y1[j] = float((j+1))
            if n < 0:
                return m * input_row_count
            return n

    if num_columns == 2:
        @omnisci('int32(int32, Column<double>, RowMultiplier,'
                 ' OutputColumn<double>, OutputColumn<double>)')
        def ret_2_columns(n, x1, m, y1, y2):
            input_row_count = len(x1)
            for i in range(input_row_count):
                for c in range(m):
                    j = i + c * input_row_count
                    y1[j] = float((j+1))
                    y2[j] = float(2*(j+1))
            if n < 0:
                return m * input_row_count
            return n

    if num_columns == 3:
        @omnisci('int32(int32, Column<double>, RowMultiplier,'
                 ' OutputColumn<double>, OutputColumn<double>,'
                 ' OutputColumn<double>)')
        def ret_3_columns(n, x1, m, y1, y2, y3):
            input_row_count = len(x1)
            for i in range(input_row_count):
                for c in range(m):
                    j = i + c * input_row_count
                    y1[j] = float((j+1))
                    y2[j] = float(2*(j+1))
                    y3[j] = float(3*(j+1))
            if n < 0:
                return m * input_row_count
            return n

    if num_columns == 4:
        @omnisci('int32(int32, Column<double>, RowMultiplier,'
                 ' OutputColumn<double>, OutputColumn<double>,'
                 ' OutputColumn<double>, OutputColumn<double>)')
        def ret_4_columns(n, x1, m, y1, y2, y3, y4):
            input_row_count = len(x1)
            for i in range(input_row_count):
                for c in range(m):
                    j = i + c * input_row_count
                    y1[j] = float((j+1))
                    y2[j] = float(2*(j+1))
                    y3[j] = float(3*(j+1))
                    y4[j] = float(4*(j+1))
            if n < 0:
                return m * input_row_count
            return n

    descr, result = omnisci.sql_execute(
        f'select * from table(ret_{num_columns}_columns({max_n}, '
        f'cursor(select f8 from {omnisci.table_name}), {sizer}));')

    result = list(result)

    if max_n == -1:
        assert len(result) == sizer * 5
    else:
        assert len(result) == max_n
    for i, r in enumerate(result):
        for j, v in enumerate(r):
            assert v == float((i+1) * (j+1))


@pytest.mark.parametrize("variant", [1, 2, 3])
def test_issue173(omnisci, variant):
    omnisci.reset()
    omnisci.register()

    def mask_zero(x, b, m, y):
        for i in range(len(x)):
            if b[i]:
                y[i] = 0.0
            else:
                y[i] = x[i]
        return len(x)

    descr, result = omnisci.sql_execute(
        f'select f8, b from {omnisci.table_name}')
    f8, b = zip(*list(result))

    # remove after resolving rbc issue 175:
    mask_zero.__name__ += f'_{variant}'

    if variant == 1:
        omnisci('int32(Column<double>, Column<bool>, RowMultiplier,'
                ' OutputColumn<double>)')(mask_zero)
        descr, result = omnisci.sql_execute(
            f'select * from table({mask_zero.__name__}('
            f'cursor(select f8 from {omnisci.table_name}),'
            f'cursor(select b from {omnisci.table_name}), 1))')
    elif variant == 2:
        omnisci('int32(Cursor<Column<double>, Column<bool>>, RowMultiplier,'
                ' OutputColumn<double>)')(mask_zero)

        descr, result = omnisci.sql_execute(
            f'select * from table({mask_zero.__name__}('
            f'cursor(select f8, b from {omnisci.table_name}), 1))')
    elif variant == 3:
        omnisci('int32(Cursor<double, bool>, RowMultiplier,'
                ' OutputColumn<double>)')(mask_zero)

        descr, result = omnisci.sql_execute(
            f'select * from table({mask_zero.__name__}('
            f'cursor(select f8, b from {omnisci.table_name}), 1))')
    else:
        raise ValueError(variant)

    result = list(result)

    for i in range(len(result)):
        assert result[i][0] == (0.0 if b[i] else f8[i])


def test_redefine(omnisci):
    omnisci.reset()
    # register an empty set of UDFs in order to avoid unregistering
    # UDFs created directly from LLVM IR strings when executing SQL
    # queries:
    omnisci.register()

    descr, result = omnisci.sql_execute(
        f'select f8, i4 from {omnisci.table_name}')

    f8, i4 = zip(*result)

    @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<double>)')  # noqa: E501, F811
    def redefined_udtf(x, m, y):  # noqa: E501, F811
        for i in range(len(x)):
            y[i] = x[i] + 1
        return len(x)

    descr, result = omnisci.sql_execute(
        'select * from table(redefined_udtf(cursor('
        f'select f8 from {omnisci.table_name}), 1))')
    result = list(result)
    for i in range(len(result)):
        assert result[i][0] == f8[i] + 1

    # redefine implementation
    @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<double>)')  # noqa: E501, F811
    def redefined_udtf(x, m, y):  # noqa: E501, F811
        for i in range(len(x)):
            y[i] = x[i] + 2
        return len(x)

    descr, result = omnisci.sql_execute(
        'select * from table(redefined_udtf(cursor('
        f'select f8 from {omnisci.table_name}), 1))')
    result = list(result)
    for i in range(len(result)):
        assert result[i][0] == f8[i] + 2

    # overload with another type
    @omnisci('int32(Column<int32>, RowMultiplier, OutputColumn<int32>)')  # noqa: E501, F811
    def redefined_udtf(x, m, y):  # noqa: E501, F811
        for i in range(len(x)):
            y[i] = x[i] + 3
        return len(x)

    descr, result = omnisci.sql_execute(
        'select * from table(redefined_udtf(cursor('
        f'select i4 from {omnisci.table_name}), 1))')
    result = list(result)
    for i in range(len(result)):
        assert result[i][0] == i4[i] + 3

    # check overload
    descr, result = omnisci.sql_execute(
        'select * from table(redefined_udtf(cursor('
        f'select f8 from {omnisci.table_name}), 1))')
    result = list(result)
    for i in range(len(result)):
        assert result[i][0] == f8[i] + 2


@pytest.mark.parametrize("step", [1, 2, 3])
def test_overload_nonuniform(omnisci, step):
    pytest.xfail('Test failing due to the introduction of default sizer. See PR 313')

    omnisci.reset()
    omnisci.register()

    if step > 0:
        @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<int64>)')  # noqa: E501, F811
        def overloaded_udtf(x, m, y):  # noqa: E501, F811
            y[0] = 64
            return 1

    if step > 1:
        @omnisci('int32(Column<float>, RowMultiplier, OutputColumn<int64>)')  # noqa: E501, F811
        def overloaded_udtf(x, m, y):  # noqa: E501, F811
            y[0] = 32
            return 1

    if step > 2:
        @omnisci('int32(Cursor<Column<float>, Column<float>>, RowMultiplier, OutputColumn<int64>)')  # noqa: E501, F811
        def overloaded_udtf(x1, x2, m, y):  # noqa: E501, F811
            y[0] = 132
            return 1

    sql_query = ('select * from table(overloaded_udtf(cursor('
                 f'select f8 from {omnisci.table_name}), 1))')
    if step > 0:
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)
        assert result[0][0] == 64

    sql_query = ('select * from table(overloaded_udtf(cursor('
                 f'select f4 from {omnisci.table_name}), 1))')
    if step > 1:
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)
        assert result[0][0] == 32
    else:
        with pytest.raises(
                Exception,
                match=(r".*(Function overloaded_udtf\(COLUMN<FLOAT>, INTEGER\) not supported"
                       r"|Could not bind overloaded_udtf\(COLUMN<FLOAT>, INTEGER\))")):
            descr, result = omnisci.sql_execute(sql_query)

    sql_query = ('select * from table(overloaded_udtf(cursor('
                 f'select f4, f4 from {omnisci.table_name}), 1))')
    if step > 2:
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)
        assert result[0][0] == 132
    else:
        with pytest.raises(
                Exception,
                match=(r".*(Function overloaded_udtf\(COLUMN<FLOAT>,"
                       r" COLUMN<FLOAT>, INTEGER\) not supported|Could not bind"
                       r" overloaded_udtf\(COLUMN<FLOAT>, COLUMN<FLOAT>, INTEGER\))")):
            descr, result = omnisci.sql_execute(sql_query)


def test_overload_uniform(omnisci):
    omnisci.reset()
    omnisci.register()

    @omnisci('int32(Column<T>, RowMultiplier, OutputColumn<T>)',
             T=['double', 'float', 'int64', 'int32', 'int16', 'int8'])
    def mycopy(x, m, y):  # noqa: E501, F811
        for i in range(len(x)):
            y[i] = x[i]
        return len(x)

    for colname in ['f8', 'f4', 'i8', 'i4', 'i2', 'i1', 'b']:
        sql_query = (f'select {colname} from {omnisci.table_name}')
        descr, result = omnisci.sql_execute(sql_query)
        expected = list(result)
        sql_query = ('select * from table(mycopy(cursor('
                     f'select {colname} from {omnisci.table_name}), 1))')
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)
        assert result == expected


omnisci_aggregators = [
    'avg', 'min', 'max', 'sum', 'count',
    'approx_count_distinct', 'sample', 'single_value', 'stddev',
    'stddev_pop', 'stddev_samp', 'variance', 'var_pop', 'var_samp',
]
omnisci_aggregators2 = ['correlation', 'corr', 'covar_pop', 'covar_samp']
# TODO: round_to_digit, mod, truncate
omnisci_functions = ['sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'abs', 'ceil', 'degrees',
                     'exp', 'floor', 'ln', 'log', 'log10', 'radians', 'round', 'sign', 'sqrt']
omnisci_functions2 = ['atan2', 'power']
omnisci_unary_operations = ['+', '-']
omnisci_binary_operations = ['+', '-', '*', '/']


@pytest.mark.parametrize("prop", ['', 'groupby'])
@pytest.mark.parametrize("oper", omnisci_aggregators + omnisci_aggregators2)
def test_column_aggregate(omnisci, prop, oper):
    omnisci.reset()
    omnisci.register()

    if prop == 'groupby':
        pytest.skip('omniscidb server crashes')

    if oper == 'single_value':
        if prop:
            return

        @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<double>)')
        def test_rbc_last(x, m, y):
            y[0] = x[len(x)-1]
            return 1

        sql_query = (f'select f8 from {omnisci.table_name}')
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)
        expected_result = result[-1:]

        sql_query = (f'select {oper}(out0) from table(test_rbc_last(cursor('
                     f'select f8 from {omnisci.table_name}), 1))')
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)

        assert result == expected_result
        return

    @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<double>)')
    def test_rbc_mycopy(x, m, y):
        for i in range(len(x)):
            y[i] = x[i]
        return len(x)

    @omnisci('int32(Cursor<double, double>, RowMultiplier,'
             ' OutputColumn<double>, OutputColumn<double>)')
    def test_rbc_mycopy2(x1, x2, m, y1, y2):
        for i in range(len(x1)):
            y1[i] = x1[i]
            y2[i] = x2[i]
        return len(x1)

    @omnisci('int32(Cursor<double, bool>, RowMultiplier,'
             ' OutputColumn<double>, OutputColumn<bool>)')
    def test_rbc_mycopy2b(x1, x2, m, y1, y2):
        for i in range(len(x1)):
            y1[i] = x1[i]
            y2[i] = x2[i]
        return len(x1)

    @omnisci('int32(Cursor<double, double, bool>, RowMultiplier,'
             'OutputColumn<double>, OutputColumn<double>, OutputColumn<bool>)')
    def test_rbc_mycopy3(x1, x2, x3, m, y1, y2, y3):
        for i in range(len(x1)):
            y1[i] = x1[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
        return len(x1)

    extra_args = ''
    if oper == 'approx_count_distinct':
        extra_args = ', 1'

    mycopy = 'test_rbc_mycopy'

    if oper in omnisci_aggregators2:
        if prop == 'groupby':
            sql_query_expected = (
                f'select {oper}(f8, d) from {omnisci.table_name} group by b')
            sql_query = (
                f'select {oper}(out0, out1) from table({mycopy}3(cursor('
                f'select f8, d, b from {omnisci.table_name}), 1)) group by out2')
        else:
            sql_query_expected = (f'select {oper}(f8, d) from {omnisci.table_name}')
            sql_query = (
                f'select {oper}(out0, out1) from table({mycopy}2(cursor('
                f'select f8, d from {omnisci.table_name}), 1))')
    else:
        if prop == 'groupby':
            sql_query_expected = (
                f'select {oper}(f8{extra_args}) from {omnisci.table_name} group by b')
            sql_query = (
                f'select {oper}(out0) from table({mycopy}2b(cursor('
                f'select f8, b from {omnisci.table_name}), 1)) group by out1')
        else:
            sql_query_expected = (f'select {oper}(f8{extra_args}) from {omnisci.table_name}')
            sql_query = (
                f'select {oper}(out0) from table({mycopy}(cursor('
                f'select f8 from {omnisci.table_name}), 1))')

    descr, result_expected = omnisci.sql_execute(sql_query_expected)
    result_expected = list(result_expected)
    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    assert result == result_expected


@pytest.mark.parametrize("oper", omnisci_functions + omnisci_functions2)
def test_column_function(omnisci, oper):
    omnisci.reset()
    omnisci.register()

    @omnisci('int32(Column<double>, RowMultiplier, OutputColumn<double>)')
    def test_rbc_mycopy(x, m, y):
        for i in range(len(x)):
            y[i] = x[i]
        return len(x)

    @omnisci('int32(Cursor<double, double>, RowMultiplier,'
             ' OutputColumn<double>, OutputColumn<double>)')
    def test_rbc_mycopy2(x1, x2, m, y1, y2):
        for i in range(len(x1)):
            y1[i] = x1[i]
            y2[i] = x2[i]
        return len(x1)

    mycopy = 'test_rbc_mycopy'

    if oper in omnisci_functions2:
        sql_query_expected = (f'select {oper}(f8, d) from {omnisci.table_name}')
        sql_query = (
            f'select {oper}(out0, out1) from table({mycopy}2(cursor('
            f'select f8, d from {omnisci.table_name}), 1))')
    else:
        sql_query_expected = (f'select {oper}(f8) from {omnisci.table_name}')
        sql_query = (
            f'select {oper}(out0) from table({mycopy}(cursor('
            f'select f8 from {omnisci.table_name}), 1))')

    descr, result_expected = omnisci.sql_execute(sql_query_expected)
    result_expected = list(result_expected)

    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    if oper in ['asin', 'acos']:
        # result contains nan
        assert repr(result) == repr(result_expected)
    else:
        assert result == result_expected


@pytest.mark.parametrize("oper", omnisci_binary_operations)
def test_column_binary_operation(omnisci, oper):
    omnisci.reset()
    omnisci.register()

    @omnisci('int32(Cursor<double, double>, RowMultiplier,'
             ' OutputColumn<double>, OutputColumn<double>)')
    def test_rbc_mycopy2(x1, x2, m, y1, y2):
        for i in range(len(x1)):
            y1[i] = x1[i]
            y2[i] = x2[i]
        return len(x1)

    mycopy = 'test_rbc_mycopy'

    sql_query_expected = (f'select f8 {oper} d from {omnisci.table_name}')
    sql_query = (
        f'select out0 {oper} out1 from table({mycopy}2(cursor('
        f'select f8, d from {omnisci.table_name}), 1))')

    descr, result_expected = omnisci.sql_execute(sql_query_expected)
    result_expected = list(result_expected)

    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    assert result == result_expected


@pytest.mark.parametrize("oper", omnisci_unary_operations)
def test_column_unary_operation(omnisci, oper):
    omnisci.reset()
    omnisci.register()

    @omnisci('int32(Cursor<double>, RowMultiplier,'
             ' OutputColumn<double>)')
    def test_rbc_mycopy(x1, m, y1):
        for i in range(len(x1)):
            y1[i] = x1[i]
        return len(x1)

    mycopy = 'test_rbc_mycopy'

    sql_query_expected = (f'select {oper} f8 from {omnisci.table_name}')
    sql_query = (
        f'select {oper} out0 from table({mycopy}(cursor('
        f'select f8 from {omnisci.table_name}), 1))')

    descr, result_expected = omnisci.sql_execute(sql_query_expected)
    result_expected = list(result_expected)

    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    assert result == result_expected


def test_create_as(omnisci):
    omnisci.reset()
    omnisci.register()

    @omnisci('int32(Cursor<double>, RowMultiplier,'
             ' OutputColumn<double>)')
    def test_rbc_mycopy(x1, m, y1):
        for i in range(len(x1)):
            y1[i] = x1[i]
        return len(x1)

    mycopy = 'test_rbc_mycopy'

    sql_query_expected = (f'select f8 from {omnisci.table_name}')
    sql_queries = [
        f'DROP TABLE IF EXISTS {omnisci.table_name}_f8',
        (f'CREATE TABLE {omnisci.table_name}_f8 AS SELECT out0 FROM TABLE({mycopy}(CURSOR('
         f'SELECT f8 FROM {omnisci.table_name}), 1))'),
        f'SELECT * FROM {omnisci.table_name}_f8']

    descr, result_expected = omnisci.sql_execute(sql_query_expected)
    result_expected = list(result_expected)

    for sql_query in sql_queries:
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)

    assert result == result_expected


@pytest.fixture(scope='function')
def create_columns(omnisci):

    # delete tables
    omnisci.sql_execute('DROP TABLE IF EXISTS datatable;')
    omnisci.sql_execute('DROP TABLE IF EXISTS kerneltable;')
    # create tables
    omnisci.sql_execute('CREATE TABLE IF NOT EXISTS datatable (x DOUBLE);')
    omnisci.sql_execute('CREATE TABLE IF NOT EXISTS kerneltable (kernel DOUBLE);')
    # add data
    omnisci.load_table_columnar('datatable', **{'x': [1.0, 2.0, 3.0, 4.0, 5.0]})
    omnisci.load_table_columnar('kerneltable', **{'kernel': [10.0, 20.0, 30.0]})
    yield omnisci
    # delete tables
    omnisci.sql_execute('DROP TABLE IF EXISTS datatable;')
    omnisci.sql_execute('DROP TABLE IF EXISTS kerneltable;')


@pytest.mark.usefixtures('create_columns')
def test_column_different_sizes(omnisci):

    if omnisci.has_cuda and omnisci.version[:2] == (5, 10):
        pytest.xfail('Different result')

    @omnisci('int32(Column<double>, Column<double>, RowMultiplier, OutputColumn<double>)')
    def convolve(x, kernel, m, y):
        for i in range(len(y)):
            y[i] = 0.0
        for i in range(len(x)):
            for j in range(len(kernel)):
                k = i + j
                if (k < len(x)):
                    y[k] += kernel[j] * x[k]
        # output has the same size as @x
        return len(x)

    _, result = omnisci.sql_execute(
        'select * from table('
        'convolve(cursor(select x from datatable),'
        'cursor(select kernel from kerneltable), 1))')

    expected = [(10.0,), (60.0,), (180.0,), (240.0,), (300.0,)]
    assert list(result) == expected


def test_column_dtype(omnisci):
    from numba import types
    table = omnisci.table_name

    @omnisci('int32(Column<T>, RowMultiplier, OutputColumn<T>)',
             T=['int32', 'int64', 'float', 'double'])
    def col_dtype_fn(x, m, y):
        sz = len(x)
        for i in range(sz):
            if x.dtype == types.int32:
                y[i] = 1
            elif x.dtype == types.int64:
                y[i] = 2
            elif x.dtype == types.float32:
                y[i] = 3.0
            elif x.dtype == types.float64:
                y[i] = 4.0
            else:
                y[i] = 'hello world'  # this ought be removed by DCE
        return sz

    inpts = (
        ('i4', 1),
        ('i8', 2),
        ('f4', 3.0),
        ('f8', 4.0),
    )
    for col, r in inpts:
        query = f'select * from table(col_dtype_fn(cursor(select {col} from {table}), 1))'
        _, result = omnisci.sql_execute(query)
        assert list(result) == [(r,)] * 5


def test_column_enumerate(omnisci):
    from rbc.externals.omniscidb import set_output_row_size

    @omnisci('int32(Column<int32>, OutputColumn<int32>)')
    def col_enumerate(x, y):
        sz = len(x)
        set_output_row_size(sz)
        for i, e in enumerate(x):
            y[i] = e
        return sz

    _, result = omnisci.sql_execute(
        f'select * from table(col_enumerate(cursor(select i4 from {omnisci.table_name})))')
    _, expected_result = omnisci.sql_execute(
        f'select rowid, i4 from {omnisci.table_name} order by rowid;')
    for (r,), (_, e) in zip(list(result), list(expected_result)):
        assert r == e



"""
CFFI builder for _rbclib.

This is meant to be listed inside cffi_modules=[...] in setup.py.

To rebuild, run setup.py develop or equivalent.
"""

from cffi import FFI
ffibuilder = FFI()

# rbclib defines two kind of functions:
#
#  1. functions written in C: these are implemented in _rbclib.c and exposed
#     to CFFI by calling ffibuilder.cdef()
#
#  2. functions written in Python: these are exposed to CFFI by using
#     declaring them as extern "C+Python" in ffibuilder.cdef().  CFFI
#     generates a C stub which can be called from C and from JIT-generated
#     code, and which in turns call the Python function which is defined using
#     @ffi.def_extern.
#
# Moreover, we need to take extra care to support Windows.  Contrarily to
# Unix-like systems, Windows requires symbols to be explicitly exported in
# order to be visible in the generated DLL/.pyd file. RBC_DLLEXPORT


ffibuilder.cdef("""
int64_t _rbclib_add_ints(int64_t a, int64_t b);
int8_t* rbclib_allocate_varlen_buffer(int64_t element_count, int64_t element_size);
void rbclib_free_buffer(int8_t *addr);

extern "C+Python" {
    int8_t* rbclib_tracing_allocate_varlen_buffer(int64_t element_count,
                                                  int64_t element_size);
    void rbclib_tracing_free_buffer(int8_t *addr);
}
""")

ffibuilder.set_source(
    "rbc.rbclib._rbclib",
    source='#include "_rbclib.h"',
    include_dirs=['rbc/rbclib'],
    sources=['rbc/rbclib/_rbclib.c'],
)



from llvmlite import ir
from numba.core import extending
from numba.core import types as nb_types
from numba.core.errors import TypingError
from rbc import irutils


@extending.intrinsic
def add_ints(typingctx, a_type, b_type):
    if (a_type, b_type) != (nb_types.int64, nb_types.int64):
        raise TypingError('add_ints(i64, i64)')

    sig = nb_types.int64(nb_types.int64, nb_types.int64)

    def codegen(context, builder, signature, args):
        assert len(args) == 2
        arg_a, arg_b = args
        int64_t = ir.IntType(64)
        fntype = ir.FunctionType(int64_t, [int64_t, int64_t])
        fn = irutils.get_or_insert_function(builder.module, fntype,
                                            name="_rbclib_add_ints")
        return builder.call(fn, [arg_a, arg_b])

    return sig, codegen



import pytest
from rbc.errors import NumbaTypeError
from rbc.tests import omnisci_fixture, sql_execute
from rbc.externals.omniscidb import table_function_error
import numpy as np


@pytest.fixture(scope='module')
def omnisci():
    for o in omnisci_fixture(globals()):
        define(o)
        yield o


def define(omnisci):

    T = ['int64', 'int32', 'int16', 'int8', 'float64', 'float32']

    @omnisci('int32(Cursor<int64, T>, T, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<T>)', T=T)
    def sqlmultiply(rowid, col, alpha, row_multiplier, rowid_out, out):
        for i in range(len(col)):
            j = rowid[i]
            out[j] = col[i] * alpha
            rowid_out[j] = i
        return len(col)

    @omnisci('int32(Cursor<int64, T>, T, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<T>)', T=T)
    def sqladd(rowid, col, alpha, row_multiplier, rowid_out, out):
        for i in range(len(col)):
            j = rowid[i]
            out[j] = col[i] + alpha
            rowid_out[j] = i
        return len(col)

    @omnisci('int32(Cursor<int64, T>, Cursor<int64, T>, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<T>)', T=T)
    def sqladd2(rowid1, col1, rowid2, col2, row_multiplier, rowid_out, out):
        for i1 in range(len(col1)):
            j1 = rowid1[i1]
            for i2 in range(len(col2)):
                j2 = rowid2[i2]
                if j1 == j2:
                    out[j1] = col1[i1] + col2[i2]
                    rowid_out[j1] = i1
                    break
        return len(col1)

    @omnisci('int32(Cursor<int64, T>, Cursor<int64, T>, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<T>)', T=T)
    def sqlmultiply2(rowid1, col1, rowid2, col2, row_multiplier, rowid_out, out):
        for i1 in range(len(col1)):
            j1 = rowid1[i1]
            for i2 in range(len(col2)):
                j2 = rowid2[i2]
                if j1 == j2:
                    out[j1] = col1[i1] * col2[i2]
                    rowid_out[j1] = i1
                    break
        return len(col1)


@pytest.mark.parametrize("kind", ['i8', 'i4', 'i2', 'i1', 'f8', 'f4'])
def test_composition(omnisci, kind):
    omnisci.require_version((5, 6), 'Requires omniscidb-internal PR 5440')

    def tonp(query):
        _, result = omnisci.sql_execute(query)
        return tuple(np.array(a) for a in zip(*list(result)))

    sqltyp = dict(i8='BIGINT', i4='INT', i2='SMALLINT', i1='TINYINT',
                  f8='DOUBLE', f4='FLOAT')[kind]

    class Algebra:

        def __init__(self, arr, select):
            self.arr = arr
            self.select = select

        def __add__(self, other):
            if isinstance(other, (int, float)):
                return type(self)(
                    self.arr + other,
                    'select out0, out1 from table(sqladd(cursor('
                    f'{self.select}), cast({other} as {sqltyp}), 1))')
            elif isinstance(other, type(self)):
                return type(self)(
                    self.arr + other.arr,
                    'select out0, out1 from table(sqladd2(cursor('
                    f'{self.select}), cursor({other.select}), 1))')
            return NotImplemented

        def __mul__(self, other):
            if isinstance(other, (int, float)):
                return type(self)(
                    self.arr * other,
                    'select out0, out1 from table(sqlmultiply(cursor('
                    f'{self.select}), cast({other} as {sqltyp}), 1))')
            elif isinstance(other, type(self)):
                return type(self)(
                    self.arr * other.arr,
                    'select out0, out1 from table(sqlmultiply2(cursor('
                    f'{self.select}), cursor({other.select}), 1))')
            return NotImplemented

        def isok(self):
            return (tonp(self.select)[1] == self.arr).all()

        def __repr__(self):
            return f'{type(self).__name__}({self.arr!r}, {self.select!r})'

    select0 = f'select rowid, {kind} from {omnisci.table_name}'
    arr0 = tonp(select0)[1]

    a = Algebra(arr0, select0)

    assert a.isok()

    assert (a + 2).isok()
    assert ((a + 2) + 3).isok()

    assert (a * 2).isok()
    assert ((a * 2) + 3).isok()

    assert (a + a).isok()
    assert (a * a).isok()

    assert ((a + 2) * a).isok()
    assert ((a + 2) * a + 3).isok()


def test_table_function_manager(omnisci):
    omnisci.require_version((5, 9), 'Requires omniscidb-internal PR 6035')

    @omnisci('int32(TableFunctionManager, Column<double>, OutputColumn<double>)')
    def my_manager_error(mgr, col, out):
        return mgr.error_message("TableFunctionManager error_message!")

    @omnisci('int32(TableFunctionManager, Column<double>, OutputColumn<double>)')
    def my_manager_row_size(mgr, col, out):
        size = len(col)
        mgr.set_output_row_size(size)
        for i in range(size):
            out[i] = col[i]
        return size

    with pytest.raises(omnisci.thrift_client.thrift.TMapDException) as exc:
        omnisci.sql_execute(
            f'select out0 from table(my_manager_error('
            f'cursor(select f8 from {omnisci.table_name})));')

    assert exc.match('Error executing table function: TableFunctionManager error_message!')

    _, result = omnisci.sql_execute(
        f'select out0 from table(my_manager_row_size('
        f'cursor(select f8 from {omnisci.table_name})));')

    expected = [(0.0,), (1.0,), (2.0,), (3.0,), (4.0,)]
    assert(list(result) == expected)


@pytest.mark.parametrize("sleep", ['ct_sleep1', 'ct_sleep2'])
@pytest.mark.parametrize("mode", [0, 1, 2, 3, 4])
def test_parallel_execution(omnisci, sleep, mode):
    """This test is affected by omniscidb server option --num-executors.
    Here we start a number of identical tasks at the same time and
    observe if the tasks are run in-parallel (ct_sleep2) or not
    (ct_sleep1).

    """
    omnisci.require_version((5, 9), 'Requires omniscidb-internal PR 5901')
    from multiprocessing import Process, Array

    def func(seconds, mode, a):
        try:
            descr, result = sql_execute(
                f'select * from table({sleep}({seconds}, {mode}));')
        except Exception as msg:
            code = -1  # error code, positive if exception is expected
            if ((mode == 2 or (mode == 3 and sleep == 'ct_sleep1')
                 and 'unspecified output columns row size' in str(msg))):
                code = 2
            elif mode == 3 and 'uninitialized TableFunctionManager singleton' in str(msg):
                code = 3
            elif mode == 4 and 'unexpected mode' in str(msg):
                code = 4
            else:
                print(msg)
            a[0] = -1    # indicates exception
            a[1] = code
            return
        for i, v in enumerate(result):
            a[i] = v[0]

    # mode > 1 correspond to exception tests. Increasing nof_jobs for
    # such tests and rerunning pytest many times may lead to omniscidb
    # server crash if num-executors>1. This may indicate that
    # omniscidb error handling is not completely thread-save.
    nof_jobs = 5 if mode <= 1 else 1
    sleep_seconds = 1

    # Initialize tasks
    processes_outputs = []
    for i in range(nof_jobs):
        a = Array('i', [0, 0, 0])
        p = Process(target=func, args=(sleep_seconds, mode, a))
        processes_outputs.append((p, a))

    # Start tasks
    for p, _ in processes_outputs:
        p.start()

    # Collect outputs
    outputs = []
    for p, a in processes_outputs:
        p.join()
        if a[0] == -1:
            if a[1] < 0:
                raise ValueError(f'unexpected failure: code={a[1]}')
            continue
        outputs.append(a[:])

    if outputs:
        outputs = sorted(outputs)
        origin = outputs[0][0]  # in milliseconds
        print('\nthread id:---> time, 1 character == 100 ms, `*` marks an UDTF execution')
        for start, end, thread_id in outputs:
            print(f'{thread_id:8x} :' + ' ' * ((start - origin) // 100)
                  + '*' * ((end - start) // 100))


def test_table_function_error(omnisci):
    omnisci.require_version((5, 8), 'Requires omniscidb-internal PR 5879')
    omnisci.reset()

    @omnisci('int32(Column<double>, double, RowMultiplier, OutputColumn<double>)')
    def my_divide(column, k, row_multiplier, out):
        if k == 0:
            return table_function_error('division by zero')
        for i in range(len(column)):
            out[i] = column[i] / k
        return len(column)

    descr, result = omnisci.sql_execute(f"""
        select *
        from table(
            my_divide(CURSOR(SELECT f8 FROM {omnisci.table_name}), 2)
        );
    """)
    assert list(result) == [(0.0,), (0.5,), (1.0,), (1.5,), (2.0,)]

    with pytest.raises(omnisci.thrift_client.thrift.TMapDException) as exc:
        omnisci.sql_execute(f"""
            select *
            from table(
                my_divide(CURSOR(SELECT f8 FROM {omnisci.table_name}), 0)
            );
        """)
    assert exc.match('Error executing table function: division by zero')


def test_raise_error(omnisci):
    omnisci.require_version((5, 8), 'Requires omniscidb-internal PR 5879')
    omnisci.reset()

    with pytest.raises(NumbaTypeError) as exc:
        @omnisci('int32(Column<double>, double, RowMultiplier, OutputColumn<double>)')
        def my_divide(column, k, row_multiplier, out):
            if k == 0:
                raise ValueError('division by zero')
            for i in range(len(column)):
                out[i] = column[i] / k
            return len(column)

        omnisci.register()

    assert exc.match('raise statement is not supported')
    omnisci.reset()


def test_issue_235(omnisci):

    @omnisci('int32(Column<int64>, RowMultiplier, OutputColumn<int64>)')
    def text_rbc_copy_rowmul(x, m, y):
        for i in range(len(x)):
            y[i] = x[i]
        return len(x)

    @omnisci('int32(Cursor<int64, double>, RowMultiplier, OutputColumn<int64>, OutputColumn<double>)')  # noqa: E501
    def text_rbc_copy_rowmul(x, x2, m, y, y2):  # noqa: F811
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
        return len(x)

    query = (f'select * from table(text_rbc_copy_rowmul('
             f'cursor(select i8, f8 from {omnisci.table_name}), 1));')
    _, result = omnisci.sql_execute(query)
    assert list(result) == list(zip(np.arange(5), np.arange(5, dtype=np.float64)))



import operator

from rbc.errors import NumbaTypeError
from .omnisci_buffer import BufferMeta, free_all_other_buffers
from numba.core import ir, types
from numba.core.compiler import CompilerBase, DefaultPassBuilder
from numba.core.compiler_machinery import FunctionPass, register_pass
from numba.core.untyped_passes import (IRProcessing,
                                       RewriteSemanticConstants,
                                       ReconstructSSA,
                                       DeadBranchPrune,)
from numba.core.typed_passes import PartialTypeInference, DeadCodeElimination


# Register this pass with the compiler framework, declare that it will not
# mutate the control flow graph and that it is not an analysis_only pass (it
# potentially mutates the IR).
@register_pass(mutates_CFG=False, analysis_only=False)
class AutoFreeBuffers(FunctionPass):
    """
    Black magic at work.

    The goal of this pass is to "automagically" free all the buffers which
    were allocated, apart the one which is used as a return value (if any).

    NOTE: at the moment of writing there are very few tests for this and it's
    likely that it is broken and/or does not work properly in the general
    case. [Remove this note once we are confident that it works well]
    """
    _name = "auto_free_buffers"  # the common name for the pass

    def __init__(self):
        FunctionPass.__init__(self)

    # implement method to do the work, "state" is the internal compiler
    # state from the CompilerBase instance.
    def run_pass(self, state):
        func_ir = state.func_ir  # get the FunctionIR object

        for blk in func_ir.blocks.values():
            for stmt in blk.find_insts(ir.Assign):
                if (
                    isinstance(stmt.value, ir.FreeVar)
                    and stmt.value.name in BufferMeta.class_names
                ):
                    break
            else:
                continue
            break
        else:
            return False  # one does not changes the IR

        for blk in func_ir.blocks.values():
            loc = blk.loc
            scope = blk.scope
            for ret in blk.find_insts(ir.Return):

                name = "free_all_other_buffers_fn"
                value = ir.Global(name, free_all_other_buffers, loc)
                target = scope.make_temp(loc)
                stmt = ir.Assign(value, target, loc)
                blk.insert_before_terminator(stmt)

                fn_call = ir.Expr.call(func=target, args=[ret.value], kws=(), loc=loc)
                lhs = scope.make_temp(loc)
                var = ir.Assign(fn_call, lhs, blk.loc)
                blk.insert_before_terminator(var)
                break

        return True  # we changed the IR


@register_pass(mutates_CFG=False, analysis_only=False)
class CheckRaiseStmts(FunctionPass):
    _name = "check_raise_stmts"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        func_ir = state.func_ir
        for blk in func_ir.blocks.values():
            for _raise in blk.find_insts(ir.Raise):
                msg = ('raise statement is not supported in '
                       'UDF/UDTFs. Please, use `return table_function_error(msg)` '
                       'to raise an error.')
                loc = _raise.loc
                raise NumbaTypeError(msg, loc=loc)
        return False


@register_pass(mutates_CFG=False, analysis_only=False)
class DTypeComparison(FunctionPass):
    _name = "DTypeComparison"

    def __init__(self):
        FunctionPass.__init__(self)

    def is_dtype_comparison(self, func_ir, binop):
        """ Return True if binop is a dtype comparison
        """
        def is_getattr(expr):
            return isinstance(expr, ir.Expr) and expr.op == 'getattr'

        if binop.fn != operator.eq:
            return False

        lhs = func_ir.get_definition(binop.lhs.name)
        rhs = func_ir.get_definition(binop.lhs.name)

        return (is_getattr(lhs) and lhs.attr == 'dtype') or \
               (is_getattr(rhs) and rhs.attr == 'dtype')

    def run_pass(self, state):
        # run as subpipeline
        from numba.core.compiler_machinery import PassManager
        pm = PassManager("subpipeline")
        pm.add_pass(PartialTypeInference, "performs partial type inference")
        pm.finalize()
        pm.run(state)

        mutated = False

        func_ir = state.func_ir
        for block in func_ir.blocks.values():
            for assign in block.find_insts(ir.Assign):
                binop = assign.value
                if not (isinstance(binop, ir.Expr) and binop.op == 'binop'):
                    continue
                if self.is_dtype_comparison(func_ir, binop):
                    var = func_ir.get_assignee(binop)
                    typ = state.typemap.get(var.name, None)
                    if isinstance(typ, types.BooleanLiteral):
                        loc = binop.loc
                        rhs = ir.Const(typ.literal_value, loc)
                        new_assign = ir.Assign(rhs, var, loc)

                        # replace instruction
                        block.insert_after(new_assign, assign)
                        block.remove(assign)
                        mutated = True

        if mutated:
            pm = PassManager("subpipeline")
            # rewrite consts / dead branch pruning
            pm.add_pass(DeadCodeElimination, "dead code elimination")
            pm.add_pass(RewriteSemanticConstants, "rewrite semantic constants")
            pm.add_pass(DeadBranchPrune, "dead branch pruning")
            pm.finalize()
            pm.run(state)
        return mutated


class OmnisciCompilerPipeline(CompilerBase):
    def define_pipelines(self):
        # define a new set of pipelines (just one in this case) and for ease
        # base it on an existing pipeline from the DefaultPassBuilder,
        # namely the "nopython" pipeline
        pm = DefaultPassBuilder.define_nopython_pipeline(self.state)
        # Add the new pass to run after IRProcessing
        pm.add_pass_after(AutoFreeBuffers, IRProcessing)
        pm.add_pass_after(CheckRaiseStmts, IRProcessing)
        pm.add_pass_after(DTypeComparison, ReconstructSSA)
        # finalize
        pm.finalize()
        # return as an iterable, any number of pipelines may be defined!
        return [pm]



import os
import sys
import pytest
from rbc.tests import omnisci_fixture
from rbc.utils import parse_version


@pytest.fixture(scope='module')
def omnisci():
    for o in omnisci_fixture(globals()):
        yield o


def test_python_version():
    varname = 'EXPECTED_PYTHON_VERSION'
    current = tuple(sys.version_info)
    expected = os.environ.get(varname)
    if os.environ.get('CI'):
        assert expected is not None, (varname, current)
    if expected is None:
        pytest.skip(
            f'Undefined environment variable {varname},'
            f' cannot test python version (current={".".join(map(str, current))})')
    expected = parse_version(expected)
    current_stripped = current[:len(expected)]
    assert expected == current_stripped


def test_omniscidb_version(omnisci):
    varname = 'EXPECTED_OMNISCIDB_VERSION'
    current = omnisci.version
    expected = os.environ.get(varname)
    if 'CI' in os.environ and expected is None:
        pytest.fail("OmniSciDB server is not running")
    if expected is None:
        pytest.skip(
            f'Undefined environment variable {varname},'
            f' cannot test omniscidb version (current={".".join(map(str, current))})')
    if expected == 'dev':
        assert current[:2] >= (5, 8), current  # TODO: update dev version periodically
        pytest.skip(
            f'omniscidb dev version is {".".join(map(str, current))}')
    expected = parse_version(expected)
    current_stripped = current[:len(expected)]
    assert expected == current_stripped


def test_numba_version():
    varname = 'EXPECTED_NUMBA_VERSION'
    import numba
    current = parse_version(numba.__version__)
    expected = os.environ.get(varname)
    if expected is None:
        pytest.skip(
            f'Undefined environment variable {varname},'
            f' cannot test numba version (current={".".join(map(str, current))})')
    expected = parse_version(expected)
    current_stripped = current[:len(expected)]
    assert expected == current_stripped



"""
Array API for rbc.
"""

from .datatypes import *  # noqa: F401, F403
from .constants import *  # noqa: F401, F403
from .creation_functions import *  # noqa: F401, F403
from .elementwise_functions import *  # noqa: F401, F403
from .statistical_functions import *  # noqa: F401, F403

__all__ = [s for s in dir() if not s.startswith('_')]



"""This module provides the following tools:

Intrinsics:

  sizeof(<numba.types.TypeRef|Type instance>)
  cast(<numba.types.voidptr instance>, <numba.types.TypeRef|StringLiteral instance>)

Methods:

  <numba.types.voidptr instance>.cast(<numba.types.TypeRef|StringLiteral instance>)

Casts:

  intp, voidptr to/from T*, T**, ..., T************* where T is a scalar type.
"""

__all__ = ["sizeof", "cast"]

from llvmlite import ir
from numba.core import extending, imputils, typing, typeconv
from numba.core import types as nb_types
from rbc.typesystem import Type


typing_registry = typing.templates.builtin_registry
lowering_registry = imputils.builtin_registry


@extending.intrinsic
def sizeof(typingctx, arg_type):
    """Return sizeof type."""
    if isinstance(arg_type, nb_types.TypeRef):
        arg_val_type = arg_type.key
    elif isinstance(arg_type, nb_types.Type):
        arg_val_type = arg_type
    else:
        return
    sig = nb_types.int32(arg_type)

    def codegen(context, builder, signature, args):
        value_type = context.get_value_type(arg_val_type)
        size = context.get_abi_sizeof(value_type)
        return ir.Constant(ir.IntType(32), size)

    return sig, codegen


@extending.intrinsic
def cast(typingctx, ptr, typ):
    """Cast pointer value to any pointer type."""
    if isinstance(typ, nb_types.StringLiteral):
        dtype = Type.fromstring(typ.literal_value)
    elif isinstance(typ, nb_types.TypeRef):
        dtype = Type.fromnumba(typ.key)
    else:
        return
    assert dtype.is_pointer
    sig = dtype.tonumba()(ptr, typ)

    def codegen(context, builder, signature, args):
        return builder.bitcast(args[0], dtype.tollvmir())

    return sig, codegen


@extending.overload_method(type(nb_types.voidptr), "cast")
def voidptr_cast_to_any(ptr, typ):
    """Convenience method for casting voidptr to any pointer type."""
    if isinstance(ptr, type(nb_types.voidptr)):
        if isinstance(typ, (nb_types.TypeRef, nb_types.StringLiteral)):

            def impl(ptr, typ):
                return cast(ptr, typ)

            return impl


@lowering_registry.lower_cast(nb_types.intp, nb_types.CPointer)
def impl_intp_to_T_star(context, builder, fromty, toty, value):
    return builder.inttoptr(value, Type.fromnumba(toty).tollvmir())


@lowering_registry.lower_cast(nb_types.CPointer, nb_types.intp)
@lowering_registry.lower_cast(nb_types.RawPointer, nb_types.intp)
def impl_T_star_to_intp(context, builder, fromty, toty, value):
    return builder.ptrtoint(value, ir.IntType(toty.bitwidth))


@lowering_registry.lower_cast(nb_types.CPointer, nb_types.RawPointer)
@lowering_registry.lower_cast(nb_types.RawPointer, nb_types.CPointer)
def impl_T_star_to_T_star(context, builder, fromty, toty, value):
    return builder.bitcast(value, Type.fromnumba(toty).tollvmir())


scalar_types = [
    getattr(nb_types, typename)
    for typename in [
        "int64",
        "int32",
        "int16",
        "int8",
        "uint64",
        "uint32",
        "uint16",
        "uint8",
        "float32",
        "float64",
        "boolean",
    ]
]
pointer_value_types = [nb_types.intp, nb_types.voidptr]
pointer_types = [nb_types.CPointer(s) for s in scalar_types] + [nb_types.voidptr]
for p1 in pointer_value_types:
    for p2 in pointer_types:
        # ANSI C requires 12 as a maximum supported pointer level.
        for i in range(12):
            if p1 != p2:
                typeconv.rules.default_type_manager.set_compatible(
                    p1, p2, typeconv.Conversion.safe
                )
                typeconv.rules.default_type_manager.set_compatible(
                    p2, p1, typeconv.Conversion.safe
                )
            p2 = nb_types.CPointer(p2)
